//******************************************************//
//********INCLUDES**************************************//
//******************************************************//
//OpenCV Includes
#ifdef _CH_
#pragma package <opencv>
#endif

#ifndef _EiC
#include "cv.h"
#include "highgui.h"
#include <stdio.h>
#include <ctype.h>
#endif

//VideoInput Library
#include "videoInput.h"

//My includes
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <math.h>
#include <time.h>
#include <windows.h>
#include "Serial.h"

//******************************************************//
//********CUSTOM DATA STRUCTURES************************//
//******************************************************//
//for holding info about individual circles "dots"
struct dotData
{
	double ix;
	double iy;
	double area;
	double wx;
	double wy;
	double wz;
};

//for holding info about groups of dots called targets
struct target
{
	int numDots;		//Number of Dots
	dotData dots[9];	//Individual Dot Information
	dotData group;		//Joint information
	dotData minDot;

	double angle;
	double angleoffset;

	int x1, y1, x2, y2;

	int num_idd;
	double idd[36];
	double smallest_idd;

	int ID;
	double min_error;

	target *nxt;
};

// for holding info about "classifiers" - trained targets
struct classifier
{
	int ID;
	int numDots;
	double angleoffset;

	int num_idd;
	double idd[36];
	double smallest_idd;

	classifier *nxt;
};


using namespace std;


//******************************************************//
//********ALL THE GLOBAL VARIABLES**********************//
//******************************************************//

// global images
IplImage *image = 0, *thresholded_image = 0, *grey = 0, *world_image = 0, *circle_image = 0;
IplImage *rawImage[4], *greyImage[4], *thresholdedImage[4], *thresholdedImageColor[4];

//for reading from VI (video input library)
unsigned char* new_image;	

// Camera capture numbers
int captureID[4];
int cameraID[4] = {-1, -1, -1, -1};
int num_cams_ordered = 0;
int calibrated = 0;

// Capture Source Library
videoInput* VI;

// frame capture size
int fWidth = 640, fHeight = 480;

// calibration matrices
CvMat* imageC = cvCreateMat(4,1,CV_32FC1);
CvMat* worldC = cvCreateMat(3,1,CV_32FC1);
CvMat* dotStorage = cvCreateMat(20,6,CV_32FC1);
CvMat* U, K, M;
CvMat* camCalib[4];
double world_cam1[18], world_cam2[18], world_cam3[18], world_cam4[18];

// real time parameters
double threshold = 80;
bool erode = 0;
double max_dot_distance = 6;
double area_desired = 200;
double area_threshold = 0.5;
int runMode = 0;

// mouse interrupt parameters
char* status = 0;
int flags = 0;

// file stream writers
ofstream fOut, fX, fY, fA;

// gui paramters
IplImage *gui = 0;
int guiWidth = 427, guiHeight = 480;
int thumbWidth, thumbHeight;
string message1 = "", message2 = "";
string mainWindow = "Indoor Localization System";
string vidWindow = "Capture Source";


//******************************************************//
//***This function converts to grey and binary**********//
//******************************************************//
void threshold_images()
{
	for(int camerai = 0; camerai<4; camerai++)
	{
		cvCvtColor( rawImage[camerai], greyImage[camerai], CV_BGR2GRAY );
		cvThreshold(greyImage[camerai], thresholdedImage[camerai], threshold, 255, CV_THRESH_BINARY_INV);
	}
	return;
}

//******************************************************//
//*****This function gets raw frames from Cams**********//
//******************************************************//
void get_frames()
{
	// Get frames of cameras and put into appropriate images
	for(int camerai = 0; camerai<4; camerai++)
	{
		if (cameraID[camerai] != -1)
		{
			if(VI->isFrameNew(captureID[camerai]))
			{
				new_image = VI->getPixels(captureID[camerai], false, true);
				rawImage[cameraID[camerai]]->imageData = (char*) new_image;
			}
		}
	}
	return;
}


//******************************************************//
//********This function draws the main GUI**************//
//******************************************************//
int draw_gui(int mode = 1)
{	
	//set gui to tan
	cvSet(gui, CV_RGB(245,245,220), 0);
	
	// dynamically set up size of thumbnail
	IplImage *thumbnail = 0;
	thumbWidth = (int) guiWidth/2;
	thumbHeight = (int) guiHeight/3;
	thumbnail = cvCreateImage(cvSize(thumbWidth, thumbHeight), 8, 3 );
	
	if (mode == 1)			//Raw Images from cams (color)
	{
		//camera 1 (quadrant 1)
		cvResize(rawImage[0], thumbnail);
		cvSetImageROI(gui, cvRect((int)guiWidth/2, (int)guiHeight/6, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 2 (quadrant 2)
		cvResize(rawImage[1], thumbnail);
		cvSetImageROI(gui, cvRect((int) 0, (int)guiHeight/6, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 3 (quadrant 3)
		cvResize(rawImage[2], thumbnail);
		cvSetImageROI(gui, cvRect((int)0, (int)guiHeight/2, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 4 (quadrant 4)
		cvResize(rawImage[3], thumbnail);
		cvSetImageROI(gui, cvRect((int)guiWidth/2, (int)guiHeight/2, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
	}
	else if (mode == 0)		//Thresholded B/W Images
	{
		//camera 1 (quadrant 1)
		cvCvtColor(thresholdedImage[0], thresholdedImageColor[0], CV_GRAY2BGR);
		cvResize(thresholdedImageColor[0], thumbnail);
		cvSetImageROI(gui, cvRect((int)guiWidth/2, (int)guiHeight/6, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 2 (quadrant 2)
		cvCvtColor(thresholdedImage[1], thresholdedImageColor[1], CV_GRAY2BGR);
		cvResize(thresholdedImageColor[1], thumbnail);
		cvSetImageROI(gui, cvRect((int) 0, (int)guiHeight/6, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 3 (quadrant 3)
		cvCvtColor(thresholdedImage[2], thresholdedImageColor[2], CV_GRAY2BGR);
		cvResize(thresholdedImageColor[2], thumbnail);
		cvSetImageROI(gui, cvRect((int)0, (int)guiHeight/2, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
		//camera 4 (quadrant 4)
		cvCvtColor(thresholdedImage[3], thresholdedImageColor[3], CV_GRAY2BGR);
		cvResize(thresholdedImageColor[3], thumbnail);
		cvSetImageROI(gui, cvRect((int)guiWidth/2, (int)guiHeight/2, thumbWidth, thumbHeight));
		cvCopy(thumbnail, gui);
	}

	// Put Title Text
	cvResetImageROI(gui);
	CvFont font;
	cvInitFont(&font,CV_FONT_HERSHEY_SIMPLEX|CV_FONT_ITALIC, 0.75,0.75,0,2);
	string title_string = "Indoor Localization System";
	cvPutText (gui,title_string.c_str(),cvPoint((int) guiWidth/8,(int) 20), &font, CV_RGB(0,0,0));
	cvInitFont(&font,CV_FONT_HERSHEY_SIMPLEX|CV_FONT_ITALIC, 0.5,0.5,0,1);
	title_string = "Version 1.0 3/25/2009";
	cvPutText (gui,title_string.c_str(),cvPoint((int) guiWidth/3.5,(int) 45), &font, CV_RGB(0,0,0));
	title_string = "Scott McLeod - Northwestern University";
	cvPutText (gui,title_string.c_str(),cvPoint((int) guiWidth/8,(int) 65), &font, CV_RGB(0,0,0));

	// Put Message Text
	cvPutText (gui,message1.c_str(),cvPoint((int) 10,(int) guiHeight - 50), &font, CV_RGB(0,0,255));
	cvPutText (gui,message2.c_str(),cvPoint((int) 10,(int) guiHeight - 30), &font, CV_RGB(0,0,255));
	cvReleaseImage(&thumbnail);
	return 0;
}

//******************************************************//
//********This function captues the keystrokes**********//
//****To adjust the dynamic parameters of processing****//
//******************************************************//
int process_keys()
{
	int c = cvWaitKey(10);
	if( (char) c == 27 )
	{
		// 'Esc' exits
		return (1);
	}
	if( (char) c == 13 )
	{
		// 'Return' used for setup (press enter to continue)
		return (2);
	}
	else if ( c == 'r')
	{
		// 'r' is used to restart all captures
		// this will assume the same ordering/phyiscal layout
		// of the cameras, but will disconnect the library and
		// attempt to reopen the source.

		// Used when a black image is seen from one of the cameras if the
		// capture source was not closed properly during the last run (commonly reason)
		for (int camerai = 0; camerai<4; camerai++)
		{
			VI->restartDevice(captureID[camerai]);
		}
	}
	else if ( c == 'e')
	{
		// e turns on and off erosion of the images.
		// default is not to use it, it doesn't seem particularly usefull with
		// the other patern and real time constraints
		erode = !erode;
		cout<<"Erosion is "<<erode<<endl;
	}
	else if ( (char) c == 43 || (char) c == 61)
	{
		// '+ or (shift+)'  Increases threshold for black/white
		if (threshold<255)
			threshold++;
		cout<<"Threshold: "<<threshold<<endl;
	}
	else if ( (char) c == 45)
	{
		// '-' Decreases threshold for black/white distinction
		if (threshold>0)
			threshold--;
		cout<<"Threshold: "<<threshold<<endl;
	}
	else if ( c == ']' )
	{
		// ']' Used to increase max distance between dots
		max_dot_distance = max_dot_distance + 1;
		cout<<"Max Dot Distance: "<<max_dot_distance<<endl;
	}
	else if ( c == '[' )
	{
		// '[' Used to decrease max distance between dots
		max_dot_distance = max_dot_distance - 1;
		cout<<"Max Dot Distance: "<<max_dot_distance<<endl;
	}
	else if ( c == 'x' )
	{
		// 'x' Used to incresae size of individual dot
		area_desired += 1;
		cout<<"area_desired: "<<area_desired<<endl;
	}
	else if ( c == 'z' )
	{
		// 'z' Used to decrease size of individual dot
		area_desired -= 1;
		cout<<"area_desired: "<<area_desired<<endl;
	}
	else
	{
		if ((char) c != -1)
			cout<<"Keypress: "<<c<<endl;
	}
	return 0;
}

//******************************************************//
//********This captures mouse input from the GUI********//
//******************************************************//
void on_mouse( int event, int x, int y, int flags, void* param )
{
    if( !image )
        return;

    if( image->origin )
	{
		y = image->height - y;
	}
	
	// Left click event on GUI is used to orient cameras in 
	// from physical locations to USB PORTs on the computer
	if( event == CV_EVENT_LBUTTONDOWN )
	{
		if (num_cams_ordered < 4)	// user clicking to specify camera locations
		{
			if (x>guiWidth / 2)
			{
				if (y>guiHeight / 2)
				{
					cameraID[num_cams_ordered] = 3;
					cout<<"\nQ4\n";
				}
				else
				{
					cameraID[num_cams_ordered] = 0;
					cout<<"\nQ1\n";
				}
			}
			else
			{
				if (y>guiHeight / 2)
				{
					cameraID[num_cams_ordered] = 2;
					cout<<"\nQ3\n";
				}
				else
				{
					cameraID[num_cams_ordered] = 1;
					cout<<"\nQ2\n";
				}
			}
			num_cams_ordered++;
		}
		else if (calibrated == 0)
		{
			calibrated = 1;
		}
		else if (runMode == 0)
		{
			runMode = 1;
		}
	}

	// Right mouse click on GUI used to display a world location
	// with calibrated world coordinates
	if( event == CV_EVENT_RBUTTONDOWN )
	{
		if (x>guiWidth / 2)
		{
			if (y>guiHeight / 2)
			{
				cout<<"\nQ4\n";
				cvmSet(imageC,0,0,x);
				cvmSet(imageC,1,0,y);
				cvmSet(imageC,2,0,1);
				cvmSet(imageC,3,0,1);

				cvMatMul(camCalib[3],imageC, worldC);
				double worldX = cvmGet(worldC, 0,0);
				double worldY = cvmGet(worldC, 1,0);
				double worldZ = cvmGet(worldC, 2,0);

				cout<<" X:"<<x<<" Y: "<<y<<" \tXw: "<<worldX<<" Yw: "<<worldY<<endl;
			}
			else
			{
				cameraID[num_cams_ordered] = 0;
				cout<<"\nQ1\n";
			}
		}
		else
		{
			if (y>guiHeight / 2)
			{
				cameraID[num_cams_ordered] = 2;
				cout<<"\nQ3\n";
			}
			else
			{
				cameraID[num_cams_ordered] = 1;
				cout<<"\nQ2\n";
			}
		}
		
	}
}


// Used to calibrate a set of image pixels to world coordinates
// Parameters are worldC -> world coordinates as input by the user in the command prompt
// and imageC -> images coordinates as found by the location of the calibration target
// Returns pointer to openCV matrix which is solution to sum of least squared solution
// For best fit for a simple camera model.
CvMat* generate_calib_matrix(double worldC[18], double imageC[9][3])
{
	//U is the world coordinates in 2-space
	CvMat U = cvMat(18,1,CV_32FC1, worldC);
	
	// Convert from c++ array to openCV matrix
	for(int i = 0; i<18; i++)
	{
		cvmSet(&U, i,0,worldC[i]);
	}

	//K is image coords as measured by camera
	//calculate K matrix
	CvMat* K = 0;
	K = cvCreateMat(18,11,CV_32FC1);
	int row = 0;
		
	// Convert from c++ array to openCV matrix
	for(int i = 0; i<9; i++)
	{
		double x = imageC[i][0];
		double y = imageC[i][1];
		double z = imageC[i][2];

		double subU = cvmGet(&U, i, 0);
		double subV = cvmGet(&U, i+1, 0);
 	
		cvmSet(K, row, 0,  x);
		cvmSet(K, row, 1,  y);
		cvmSet(K, row, 2,  z);
		cvmSet(K, row, 3,  1);
		cvmSet(K, row, 4,  0);
		cvmSet(K, row, 5,  0);
		cvmSet(K, row, 6,  0);
		cvmSet(K, row, 7,  0);
		cvmSet(K, row, 8,  -subU*x);
		cvmSet(K, row, 9,  -subU*y);
		cvmSet(K, row, 10, -subU*z);

		cvmSet(K, row + 1, 0,  0);
		cvmSet(K, row + 1, 1,  0);
		cvmSet(K, row + 1, 2,  0);
		cvmSet(K, row + 1, 3,  0);
		cvmSet(K, row + 1, 4,  x);
		cvmSet(K, row + 1, 5,  y);
		cvmSet(K, row + 1, 6,  z);
		cvmSet(K, row + 1, 7,  1);
		cvmSet(K, row + 1, 8,  -subV*x);
		cvmSet(K, row + 1, 9,  -subV*y);
		cvmSet(K, row + 1, 10, -subV*z);
		row+=2;
	}

	CvMat* superU = 0;
	superU = cvCloneMat(&U);
	CvMat* M = cvCreateMat(11,1,CV_32FC1);
	
	// Print U matrix
	cout<<"************* U *************"<<endl;
	for(int i = 0; i<18; i++)
	{
		cout<<cvmGet(&U, i,0)<<endl;
	}

	// Print K matrix
	cout<<"************* K *************"<<endl;
	for(int i = 0; i<18; i++)
	{
		for(int j = 0; j<11; j++)
		{
			cout<<cvmGet(K, i, j)<<" ";
		}
		cout<<endl;
	}

	// Solve SSD M matrix and display
	cout<<"************* M *************"<<endl;
	cvSolve(K, &U, M, CV_SVD);
	
	// Put this solution into a 3x4 matrix so we can multiply
	// for calibration
	CvMat* CamCalib = cvCreateMat(3,4,CV_32FC1);
	cvSetZero(CamCalib);
	cvmSet(CamCalib, 0, 0, cvmGet(M,0,0));
	cvmSet(CamCalib, 0, 1,  cvmGet(M,1,0));
	cvmSet(CamCalib, 0, 2, cvmGet(M,2,0));
	cvmSet(CamCalib, 0, 3, cvmGet(M,3,0));
	cvmSet(CamCalib, 1, 0, cvmGet(M,4,0));
	cvmSet(CamCalib, 1, 1, cvmGet(M,5,0));
	cvmSet(CamCalib, 1, 2, cvmGet(M,6,0));
	cvmSet(CamCalib, 1, 3, cvmGet(M,7,0));
	cvmSet(CamCalib, 2, 0, cvmGet(M,8,0));
	cvmSet(CamCalib, 2, 1, cvmGet(M,9,0));
	cvmSet(CamCalib, 2, 2, cvmGet(M,10,0));
	double M34 = sqrt( pow(cvmGet(CamCalib,2,0),2) + pow((cvmGet(CamCalib,2,1)),2) +pow((cvmGet(CamCalib,2,2)),2));
	cvmSet(CamCalib, 2, 3,  M34);

	// Ouput M matrix
	for(int i = 0; i<3; i++)
	{
		for(int j = 0; j<4; j++)
		{
			cout<<cvmGet(CamCalib,i,j)<<"\t";
		}
		cout<<endl;
	}
	return CamCalib;
}




//******************************************************//
//********QuickSort for ordering Low to High************//
//******************************************************//
// Used to distiguish patterns as they are sorted by
// The unique spacings between all the dots in each pattern
void q_sort(double numbers[], int left, int right)
{
   double pivot, l_hold, r_hold;
   l_hold = left;
   r_hold = right;
   pivot = numbers[left];
   while (left < right)
   {
      while ((numbers[right] >= pivot) && (left < right))
      right--;
      if (left != right)
      {
	 numbers[left] = numbers[right];
	 left++;
      }
      while ((numbers[left] <= pivot) && (left < right))
	 left++;
      if (left != right)
      {
	  numbers[right] = numbers[left];
	  right--;
      }
   }
   numbers[left] = pivot;
   pivot = left;
   left = l_hold;
   right = r_hold;
   if (left < pivot)
      q_sort(numbers, left, pivot-1);
   if (right > pivot)
      q_sort(numbers, pivot+1, right);
}


//******************************************************//
//***Reads in trained txt file to create classifiers****//
//******************************************************//
// the file "target_classifiers.txt" must be in the root directory.
// this is computed by the other VC++ solution
classifier* LoadTargetData()
{
	ifstream fin("target_classifiers.txt");
	string line;

	classifier* current = new classifier;
	classifier* first = current;
	classifier* previous = new classifier;
	while(!fin.eof())
	{
		//create a new classifier and initialize it
		classifier* new_classifier = new classifier;
		new_classifier->nxt = NULL;
		new_classifier->ID = -1;
		new_classifier->numDots = -1;
		new_classifier->angleoffset = -1;

		//read one line of data from txt file
		fin>>line;

		string data;
		int idd_num = 0;
		//process data, looking for commas
		while(line.find(",")!=string::npos)
		{
			data = line.substr(0, line.find(","));
			line = line.substr(line.find(",") + 1);
			
			if (new_classifier->ID == -1)
			{
				new_classifier->ID = atof(data.c_str());
			}
			else if (new_classifier->numDots == -1)
			{
				new_classifier->numDots= atof(data.c_str());
			}
			else if (new_classifier->angleoffset == -1)
			{
				new_classifier->angleoffset = atof(data.c_str());
			}
			else
			{
				new_classifier->idd[idd_num] = atof(data.c_str());
				idd_num++;
			}
		}
		current->num_idd = idd_num;
		previous = current;
		current->nxt = new_classifier;
		current = current->nxt;
	}
	previous->nxt = NULL;
	first = first->nxt;
	return first;
	fin.close();
}


//******************************************************//
//****************The main program loop*****************//
//******************************************************//
int main( int argc, char** argv )
{
	classifier* first_classifier;
	first_classifier = LoadTargetData();

	classifier* current_classifier;
	current_classifier = first_classifier;
	
	cout<<"Scott McLeod Indoor Localization System v1.0\nMarch 2008\n";
	cout<<"Serial Port Setup.\n";
	cout<<"Type the name of the serial port (e.g. 'COM1')\n";
	string serialName;
	cin>>serialName;
	CSerial serialout;
	cout<<"Attempting to open the serial port.\n";
	serialout.Open(serialName.c_str());
	cout<<"Testing "<<serialName<<"...\n";
	cout<<"Type 't' to test connection.  Type 'q' to stop testing"<<endl;
	char sentinel = 't';
	do
	{	
		string test = "Hello World on Port ";
		test.append(serialName);
		test.append("\n");
		serialout.Write(test.c_str());
		cin>>sentinel;
	} while (sentinel == 't');
	
	cout<<"\n\n\nOpening Video Capture Devices\n\n";
		
	VI = new videoInput();
	VI->listDevices();
	
	for(int camerai = 0; camerai <4; camerai++)
	{
		cout<<"Select Capture Device "<<camerai+1<<": ";
		cin>>cameraID[camerai];
	}
	/*captureID[0] = 1;
	captureID[1] = 2;
	captureID[2] = 3;
	captureID[3] = 4;
	*/
	
	// connect to all 4 cameras
	for (int camerai = 0; camerai<4; camerai++)
	{
		VI->setupDevice(captureID[camerai],fWidth, fHeight);
	}

	// create CV windows
	cvNamedWindow( mainWindow.c_str(), 1 );
	cvNamedWindow( vidWindow.c_str(), 1 );

	// set the mouse callback
    cvSetMouseCallback( mainWindow.c_str(), on_mouse, 0 );
	
	//initialize fps tracking vars
	// may want to change this to keep track of last 10 frames
	// so that the averaging is not so rough for longer times
	int numFrames = 0;
	int startTime = 0;
	double fps;
    
	//initialize CCL storage elements
	CvMemStorage* storage = cvCreateMemStorage(0);
    CvSeq* contour = 0;

	//initialize file writing 
	fOut.open("output.txt");
	//fX.open("X.txt");
	//fY.open("Y.txt");
	//fA.open("Angle.txt");
	
	//intialize video writers
	// These were used for saving results to video files for intermediate
	// Documentation and presentation.
	//CvVideoWriter *writer1 = 0;
	//CvVideoWriter *writer2 = 0;
	//writer1 = cvCreateVideoWriter("Contoured.avi", CV_FOURCC('D', 'I', 'V', 'X'), 29, cvSize(fWidth, fHeight), 1);
	//writer2 = cvCreateVideoWriter("Original.avi", CV_FOURCC('D', 'I', 'V', 'X'), 29, cvSize(fWidth, fHeight), 1);
	
	//initialize images
	image = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	grey = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
	thresholded_image = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
	circle_image = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
	world_image = cvCreateImage( cvSize(fWidth, fHeight), 8, 3 );
	gui = cvCreateImage(cvSize(guiWidth, guiHeight), 8, 3);
	
	// Init openCV drawing FONT
	CvFont font;
	cvInitFont(&font,CV_FONT_HERSHEY_SIMPLEX|CV_FONT_ITALIC, 3,3,0,2);
	
	// Draw 1 2 3 4 for quadrants
	string title_string = "1";
	rawImage[0] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	cvPutText (rawImage[0],title_string.c_str(),cvPoint((int) fWidth/2-20,(int) fHeight/2-20), &font, CV_RGB(0,255,255));
	title_string = "2";
	rawImage[1] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	cvPutText (rawImage[1],title_string.c_str(),cvPoint((int) fWidth/2-20,(int) fHeight/2-20), &font, CV_RGB(0,255,255));
	title_string = "3";
	rawImage[2] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	cvPutText (rawImage[2],title_string.c_str(),cvPoint((int) fWidth/2-20,(int) fHeight/2-20), &font, CV_RGB(0,255,255));
	title_string = "4";
	rawImage[3] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	cvPutText (rawImage[3],title_string.c_str(),cvPoint((int) fWidth/2-20,(int) fHeight/2-20), &font, CV_RGB(0,255,255));

	for(int camerai = 0; camerai<4; camerai++)
	{
		greyImage[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
		thresholdedImage[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
		thresholdedImageColor[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	}

	//******************************************************//
	//*****PHASE 1: CAMERA SETUP/ORDERING/POSITING LOOP*****//
	//******************************************************//
	// This is when the user sees a window with a cpature source
	// They must pair the capture source with the location in the GUI/world
	message1 = "Click Quadrant Corresponding to Camera";
	message2 = "Press 'r' to restart cameras (if no image visible)";
	while(num_cams_ordered < 4)
	{
		// Update vidWindow (window 2)
		new_image = VI->getPixels(captureID[num_cams_ordered], false, true);
		image->imageData = (char*) new_image;
		cvShowImage(vidWindow.c_str(), image);
		
		// Update main GUI
		get_frames();
		draw_gui();
		cvShowImage( mainWindow.c_str(), gui);
		
		// Allow Keystrokes
		if(process_keys()==1)
		{
			break;
		}
	}
	cvDestroyWindow(vidWindow.c_str());

	message1 = "Check to Make Sure Accurate Alignment";
	message2 = "Press 'Enter' to Continue";
	while(1)
	{
		get_frames();
		draw_gui();
		cvShowImage(mainWindow.c_str(), gui);
		// Allow Keystrokes
		if(process_keys()==2)
		{
			break;
		}
		
	}

	//******************************************************//
	//********PHASE 2a: WORLD CALIBRATION LOOP**************//
	//******************************************************//
	//Show template image, and input arguments for sizing
	IplImage *template_image = cvLoadImage("world_coord_template.jpg");
	double idd_dist, itd_hor, itd_ver;
	message1 = "Enter Template Pattern Constraints";
	message2 = "Type Parameter A in command prompt.";
	draw_gui(1);
	cvSetImageROI(gui, cvRect(0, guiHeight/6, template_image->width, template_image->height));
	cvCopy(template_image, gui);
	cvResetImageROI(gui);
	cvShowImage(mainWindow.c_str(), gui);
	cvWaitKey(100);
	cout<<"\nEnter parameter A in unitless double format: ";
	cin>>idd_dist;
	
	message2 = "Type Parameter B in command prompt.";
	draw_gui(1);
	cvSetImageROI(gui, cvRect(0, guiHeight/6, template_image->width, template_image->height));
	cvCopy(template_image, gui);
	cvResetImageROI(gui);
	cvShowImage(mainWindow.c_str(), gui);
	cvWaitKey(100);
	cout<<"\nEnter parameter B in unitless double format: ";
	cin>>itd_ver;
	
	message2 = "Type Parameter C in command prompt.";
	draw_gui(1);
	cvSetImageROI(gui, cvRect(0, guiHeight/6, template_image->width, template_image->height));
	cvCopy(template_image, gui);
	cvResetImageROI(gui);
	cvShowImage(mainWindow.c_str(), gui);
	cvWaitKey(100);
	cout<<"\nEnter parameter C in unitless double format: ";
	cin>>itd_hor;

	//gerenate the world coordinate matrices.
	// This computes the location of all 9 x 4 = 28 dots
	// Assuming they are placed in a perfectly rectangular pattern
	//set origin to lower left (Q3) bottom left "dot"
	int index = 0;
	for(int col = 0; col<18; col+=6)
	{
		for(int row = 4; row>=0; row-=2)
		{
			world_cam3[index] = idd_dist*(col/6);
			world_cam3[index+1] = -idd_dist*(4-row)/2;
			index+=2;
		}
	}
	cout<<"Q3";
	for(int i = 0; i<18; i++)
	{
		cout<<"element: "<<i<<": "<<world_cam3[i]<<endl;
	}

	//define all other cameras based on idd_hor and idd_vert
	//simply offset the 3x3 grid in the distances defined 'B'/'C'
	for (int i = 0; i<18; i+=2)
	{
		world_cam2[i] = world_cam3[i];
		world_cam2[i+1] = world_cam3[i+1]+itd_ver;

		world_cam1[i] = world_cam3[i]+itd_hor;
		world_cam1[i+1] = world_cam3[i+1]+itd_ver;

		world_cam4[i] = world_cam3[i]+itd_hor;
		world_cam4[i+1] = world_cam3[i+1];
	}

	//******************************************************//
	//********PHASE 2b: CAMERA CALIBRATION LOOP**************//
	//******************************************************//
	message1 = "Align Calibration Pattern";
	message2 = "Press 'Enter' to Calibrate to World.";
	//message2 = "Use +/- to Adjust Threshold.";
	
	for(int camerai = 0; camerai<4; camerai++)
	{
		rawImage[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
		greyImage[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
		thresholdedImage[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 1 );
		thresholdedImageColor[camerai] = cvCreateImage(cvSize(fWidth, fHeight), 8, 3 );
	}

	while(1)
	{
		// update GUI
		get_frames();
		// Threshold images
		threshold_images();
		draw_gui(0);
		cvShowImage( mainWindow.c_str(), gui);

		// Allow Keystrokes (enter to continue)
		// When the user hits enter, the process_keys() loop will break
		// and the current frames will be captured for calibration
		if(process_keys()==2)
		{
			break;
		}
	}
	
	message1 = "Calculating Camera Parameters";
	message2 = "Please wait...";
	for(int camerai = 0; camerai<4; camerai++)
	{
		// process image, find countours and dots
		cvFindContours(thresholdedImage[camerai], storage, &contour, sizeof(CvContour), CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );
		target *calib_pattern = new target;
		calib_pattern->numDots = 0;
		for(contour; contour; contour=contour->h_next)
		{
			//find moments of contours
			CvMoments moments;
			cvContourMoments(contour, &moments);
			double area = cvGetSpatialMoment(&moments, 0, 0);
			// area thresholding
			if ( (area>area_desired*(1-area_threshold)) )
			{
				double centerx = cvGetSpatialMoment(&moments, 1, 0) / area;
				double centery = cvGetSpatialMoment(&moments, 0, 1) / area;
				
				if (calib_pattern->numDots < 9)
				{
					calib_pattern->dots[calib_pattern->numDots].ix = centerx;
					calib_pattern->dots[calib_pattern->numDots].iy = centery;
					calib_pattern->numDots++;
				}
			}
		}

		// sort dots
		target *calib_pattern_sorted = new target;
		for(int round = 0; round < 3; round++)
		{
			double minX1, minX2, minX3;
			int posX1, posX2, posX3;
			minX1 = minX2 = minX3 = 100000;
			posX1 = posX2 = posX3 = 0;
			// Sort X's
			for (int i = 0; i<calib_pattern->numDots; i++)
			{
				if(calib_pattern->dots[i].ix<minX3)
				{
					if (calib_pattern->dots[i].ix<minX2)
					{
						if (calib_pattern->dots[i].ix<minX1)
						{
							minX3 = minX2;
							posX3 = posX2;
							minX2 = minX1;
							posX2 = posX1;
							minX1 = calib_pattern->dots[i].ix;
							posX1 = i;
						}
						else
						{
							minX3 = minX2;
							posX3 = posX2;
							minX2 = calib_pattern->dots[i].ix;
							posX2 = i;
						}
					}
					else
					{
						minX3 = calib_pattern->dots[i].ix;
						posX3 = i;
					}
				}
			}
			//Sort Y's
			if (calib_pattern->dots[posX1].iy < calib_pattern->dots[posX2].iy)
			{
				if (calib_pattern->dots[posX1].iy < calib_pattern->dots[posX3].iy)
				{
					calib_pattern_sorted->dots[round*3 + 0] = calib_pattern->dots[posX1];
					if (calib_pattern->dots[posX2].iy < calib_pattern->dots[posX3].iy)
					{
						calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX2];
						calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX3];
					}
					else
					{
						calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX3];
						calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX2];
					}
				}
				else
				{
					calib_pattern_sorted->dots[round*3 + 0] = calib_pattern->dots[posX3];
					calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX1];
					calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX2];
				}	
			}
			else
			{
				if (calib_pattern->dots[posX2].iy < calib_pattern->dots[posX3].iy)
				{
					calib_pattern_sorted->dots[round*3 + 0] = calib_pattern->dots[posX2];
					if (calib_pattern->dots[posX1].iy < calib_pattern->dots[posX3].iy)
					{
						calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX1];
						calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX3];
					}
					else
					{
						calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX3];
						calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX1];
					}
				}
				else
				{
					calib_pattern_sorted->dots[round*3 + 0] = calib_pattern->dots[posX3];
					calib_pattern_sorted->dots[round*3 + 1] = calib_pattern->dots[posX2];
					calib_pattern_sorted->dots[round*3 + 2] = calib_pattern->dots[posX1];
				}	
			}

			calib_pattern->dots[posX1].ix = 100000;
			calib_pattern->dots[posX2].ix = 100000;
			calib_pattern->dots[posX3].ix = 100000;
		}
		//output sorted
		/*cout<<"Camera 1 Sorted:"<<endl;
		for (int i = 0; i<9; i++)
		{
			cout<<i<<": ("<<calib_pattern_sorted->dots[i].ix<<", "<<calib_pattern_sorted->dots[i].iy<<")"<<endl;
		}*/


		//Generate calibration matrices
		double image_coords[9][3];
		for(int i = 0; i<9; i++)
		{
			image_coords[i][0] = calib_pattern_sorted->dots[i].ix;
			image_coords[i][1] = calib_pattern_sorted->dots[i].iy;
			image_coords[i][2] = 1;
		}

		double world_coords[18] = {-6.6, 6.6, 0, 6.6, 6.6, 6.6, -6.6, 0, 0, 0, 6.6, 0, -6.6, -6.6, 0, -6.6, 6.6, -6.6};

		// Call camera calibration function to get the 4 calibration matricies
		if (camerai == 0)
		{
			camCalib[camerai] = generate_calib_matrix(world_cam1, image_coords);
		}
		else if (camerai == 1)
		{
			camCalib[camerai] = generate_calib_matrix(world_cam2, image_coords);
		}
		else if (camerai == 2)
		{
			camCalib[camerai] = generate_calib_matrix(world_cam3, image_coords);
		}
		else if (camerai == 3)
		{
			camCalib[camerai] = generate_calib_matrix(world_cam4, image_coords);
		}

	}

	//******************************************************//
	//********PHASE 3a: PREP MAIN PROGRAM LOOP**************//
	//******************************************************//	
	message1 = "Remove the Calibration Patterns";
	message2 = "Press 'Enter' when ready to start";
	while(1)
	{
		// update GUI
		get_frames();
		// Threshold images
		threshold_images();
		draw_gui(0);  
		cvShowImage( mainWindow.c_str(), gui);

		// Allow Keystrokes (Enter exits)
		if(process_keys()==2)
		{
			break;
		}
	}

	//******************************************************//
	//********PHASE 3b: MAIN PROGRAM LOOP*******************//
	//******************************************************//	
	message1 = "Real-Time Localization Active. Press Esc to Quit";
	message2 = "";
	startTime = clock();
	while(1)
	{
		//Update GUI
		threshold_images();
		draw_gui(0);

		get_frames();
		threshold_images();
		draw_gui(0);
		cvShowImage( mainWindow.c_str(), gui);

		//setup temp pointer to target list
		target* targets_temp[4];
		for(int i = 0; i<4; i++)
		{
			targets_temp[i] = new target;
			targets_temp[i]->nxt = NULL;
			targets_temp[i]->ID = -1;
		}

		// setup scanning variables
		int camScan[4] = {-1, -1, -1, -1};
		int numScanned = 0;
		int camerai = 0;

		// scan all 4 cameras before processing
		while (numScanned != 4)
		{
			camerai++;
			if (camerai > 3)
			{
				camerai = 0;
			}
			// camera setup and camera's frame hasnt' been grabbed
			// if camScan == -1, then the camera hasn't been checked
			// This ensures that all 4 cameras are processed before the
			// Same camera is processed twice.  VI checks if the frames are new
			// to prevent redundant checking of the same frame.
			if (cameraID[camerai] != -1 && camScan[camerai] == -1 )
			{
				if(VI->isFrameNew(captureID[camerai]))
				{
					//mark it gave frame and increase count
					camScan[camerai] = 0;
					numScanned++;
					
					//capture image
					new_image = VI->getPixels(captureID[camerai], false, true);
					image->imageData = (char*) new_image;
					//convert to b/w and threshold
					cvCvtColor( image, grey, CV_BGR2GRAY );
					cvThreshold(grey, thresholded_image, threshold, 255, CV_THRESH_BINARY_INV);
					//find countours CCL (locate dots)
					cvFindContours(thresholded_image, storage, &contour, sizeof(CvContour), CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );
					CvScalar color1, color2;
					color1 = CV_RGB(255, 0, 0);
					color2 = CV_RGB(0, 255, 0);

					// set up moments
					int i = 0;
					double totalx = 0, totaly = 0;
					double centerCX = 0, centerCY = 0;
					
					//clear storage
					target* first_target = new target;
					first_target->numDots = 0;
					first_target->nxt = NULL;

					//for all contours, find dots and add them to storage
					CvMoments moments;
					for(contour;contour; contour=contour->h_next)
					{
						i++;	
						//find moments of contours
						cvContourMoments(contour, &moments);
						double area = cvGetSpatialMoment(&moments, 0, 0);
						
						// area thresholding
						// must be within (1-area_threshold) and (1 + area_threshold) of area_desired
						// area_desired is real time parameter adjustable with 'z' and 'x'
						// default is area_threshold = 0.5, so within 0.5 and 1.5 of area_desired
						if ( (area>area_desired*(1-area_threshold)) && (area<area_desired*(1+area_threshold)) )
						{
							double centerx = cvGetSpatialMoment(&moments, 1, 0) / area;
							double centery = cvGetSpatialMoment(&moments, 0, 1) / area;
							
							//Convert Image Coordinates to World Coordinates
							cvSetZero(imageC);
							
							//Set up image Matrix
							cvmSet(imageC,0,0,centerx);
							cvmSet(imageC,1,0,centery);
							cvmSet(imageC,2,0,1);
							cvmSet(imageC,3,0,1);

							//Multiply by Camera Calibration Matrix
							cvMatMul(camCalib[cameraID[camerai]],imageC, worldC);

							//Extract world coordinates
							double worldX = cvmGet(worldC, 0,0);
							double worldY = cvmGet(worldC, 1,0);
							double worldZ = cvmGet(worldC, 2,0);

							//check for best fit target, create new otherwise
							target* temp = first_target;
							bool found_match = false;

							if ((first_target->nxt == NULL) && (first_target->numDots == 0))
							{
								first_target->dots[0].ix = centerx;
								first_target->dots[0].iy = centery;
								first_target->dots[0].wx = worldX;
								first_target->dots[0].wy = worldY;
								first_target->dots[0].wz = worldZ;
								first_target->dots[0].area = area;
								first_target->numDots = 1;
								found_match = true;
							}

							while((temp) && (!found_match))
							{
								//look for matching dot in each group
								for (i = 0; i<temp->numDots; i++)
								{
									double dist = sqrt(pow((worldX - temp->dots[i].wx),2) + pow( (worldY - temp->dots[i].wy),2));
									// If the distance between this dot and another inside a group is small enough
									// then associate this dot with this group "target"
									if ( (dist < max_dot_distance)  && (temp->numDots<9) && (!found_match))  
									{
										temp->dots[temp->numDots].ix = centerx;
										temp->dots[temp->numDots].iy = centery;
										temp->dots[temp->numDots].wx = worldX;
										temp->dots[temp->numDots].wy = worldY;
										temp->dots[temp->numDots].wz = worldZ;
										temp->dots[temp->numDots].area = area;
										temp->numDots++;
										found_match = true;
									}
								}
								temp = temp->nxt;
							}

							//didnt' find a match, make a new target object for the dot
							if (!found_match)
							{
								target* new_target = new target;
								new_target->numDots = 1;
								new_target->nxt = NULL;

								new_target->dots[0].ix = centerx;
								new_target->dots[0].iy = centery;
								new_target->dots[0].wx = worldX;
								new_target->dots[0].wy = worldY;
								new_target->dots[0].wz = worldZ;
								new_target->dots[0].area = area;
								//find last node
								
								temp = first_target;
								while(temp->nxt != NULL)
								{
									temp = temp->nxt;
								}
								temp->nxt = new_target;
							}
						}


					}

					
					// All the dots have now been sorted into targets


					//find group center of mass
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						targets_temp[camerai]->group.wx = 0;
						targets_temp[camerai]->group.wy = 0;
						targets_temp[camerai]->group.wz = 0;
						targets_temp[camerai]->group.ix = 0;
						targets_temp[camerai]->group.iy = 0;
						targets_temp[camerai]->group.area = 0;

						for(int i = 0; i<targets_temp[camerai]->numDots; i++)
						{
							targets_temp[camerai]->group.wx += targets_temp[camerai]->dots[i].wx;
							targets_temp[camerai]->group.wy += targets_temp[camerai]->dots[i].wy;
							targets_temp[camerai]->group.wz += targets_temp[camerai]->dots[i].wz;
							targets_temp[camerai]->group.ix += targets_temp[camerai]->dots[i].ix;
							targets_temp[camerai]->group.iy += targets_temp[camerai]->dots[i].iy;
							targets_temp[camerai]->group.area += targets_temp[camerai]->dots[i].area;
						}
						targets_temp[camerai]->group.ix /= targets_temp[camerai]->numDots;
						targets_temp[camerai]->group.iy /= targets_temp[camerai]->numDots;
						targets_temp[camerai]->group.wx /= targets_temp[camerai]->numDots;
						targets_temp[camerai]->group.wy /= targets_temp[camerai]->numDots;
						targets_temp[camerai]->group.wz /= targets_temp[camerai]->numDots;
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//find bounding rect of TARGETS  (min enclosing rect that contains all points)
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						targets_temp[camerai]->x1 = targets_temp[camerai]->x2 = targets_temp[camerai]->dots[0].ix;
						targets_temp[camerai]->y1 = targets_temp[camerai]->y2 = targets_temp[camerai]->dots[0].iy;
						for(int i = 0; i<targets_temp[camerai]->numDots; i++)
						{
							if ( targets_temp[camerai]->dots[i].ix < targets_temp[camerai]->x1)
								targets_temp[camerai]->x1 = targets_temp[camerai]->dots[i].ix;
							
							if ( targets_temp[camerai]->dots[i].iy < targets_temp[camerai]->y1)
								targets_temp[camerai]->y1 = targets_temp[camerai]->dots[i].iy;
							
							if ( targets_temp[camerai]->dots[i].ix > targets_temp[camerai]->x2)
								targets_temp[camerai]->x2 = targets_temp[camerai]->dots[i].ix;
							
							if ( targets_temp[camerai]->dots[i].iy > targets_temp[camerai]->y2)
								targets_temp[camerai]->y2 = targets_temp[camerai]->dots[i].iy;
						}
						
						targets_temp[camerai]->x1 = targets_temp[camerai]->x1 - max_dot_distance;
						targets_temp[camerai]->x2 = targets_temp[camerai]->x2 + max_dot_distance;
						targets_temp[camerai]->y1 = targets_temp[camerai]->y1 - max_dot_distance;
						targets_temp[camerai]->y2 = targets_temp[camerai]->y2 + max_dot_distance;
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//Find interdottal distances
					// This is the unique distance between a set of dots
					// Will be used to match a target to a classifier
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						int count = 0;
						for(int i = 0; i<targets_temp[camerai]->numDots; i++)
						{
							for(int j = i+1; j<targets_temp[camerai]->numDots; j++)
							{
								double dist = sqrt(pow((targets_temp[camerai]->dots[i].wx - targets_temp[camerai]->dots[j].wx),2) + pow( (targets_temp[camerai]->dots[i].wy - targets_temp[camerai]->dots[j].wy),2));
								targets_temp[camerai]->idd[count] = dist;
								count++;
							}
						}
						targets_temp[camerai]->num_idd = count;
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//Find smallest idd
					// smallest "inter-dot-distance" used to normalize the distances between the dots
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						targets_temp[camerai]->smallest_idd = targets_temp[camerai]->idd[0];
						for(int i = 0; i<targets_temp[camerai]->num_idd; i++)
						{
							if (targets_temp[camerai]->idd[i]<targets_temp[camerai]->smallest_idd)
							{	
								targets_temp[camerai]->smallest_idd = targets_temp[camerai]->idd[i];
							}
						}
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//Normalize to smallest idd
					//sort distances in ascending order
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						// use the quick sort function at the top of the program
						q_sort(targets_temp[camerai]->idd, 0, targets_temp[camerai]->num_idd-1);
						//normalize
						for(int i = 0; i<targets_temp[camerai]->num_idd; i++)
						{
							targets_temp[camerai]->idd[i] /= targets_temp[camerai]->smallest_idd;
						}
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//peform comparison between IDD
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])															//for each target
					{
						current_classifier = first_classifier;
						targets_temp[camerai]->min_error = -1;
						targets_temp[camerai]->ID = -1;
						while(current_classifier)															//for each classifier (as calculated form the input file "target_classifiers.txt"
						{
							if (current_classifier->numDots == targets_temp[camerai]->numDots)					//if Number of DOTS match
							{
								double error = 0;
								for (int i = 0; i<targets_temp[camerai]->num_idd; i++)							//calculate min squared error
								{
									error += pow(current_classifier->idd[i] - targets_temp[camerai]->idd[i], 2);
								}
								if (error < targets_temp[camerai]->min_error || targets_temp[camerai]->min_error == -1)
								{
									targets_temp[camerai]->min_error = error;
									targets_temp[camerai]->ID = current_classifier->ID;
									targets_temp[camerai]->angleoffset = current_classifier->angleoffset;
								}
							}
							current_classifier = current_classifier->nxt;
						}
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//find closest point to center of mass and angle
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						double min_dist = -1;
						for(int i = 0; i<targets_temp[camerai]->numDots; i++)
						{
							double temp_dist = sqrt(pow((targets_temp[camerai]->group.wx - targets_temp[camerai]->dots[i].wx),2) + pow( (targets_temp[camerai]->group.wy  - targets_temp[camerai]->dots[i].wy),2));
							if (temp_dist < min_dist || min_dist == -1)
							{
								min_dist = temp_dist;
								targets_temp[camerai]->minDot = targets_temp[camerai]->dots[i];
							}
						}
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}


					//calculate angle
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						if (targets_temp[camerai]->minDot.wx < targets_temp[camerai]->group.wx)
						{
							if(targets_temp[camerai]->minDot.wy < targets_temp[camerai]->group.wy)				//quadrant 3				
							{
								targets_temp[camerai]->angle = atan((targets_temp[camerai]->minDot.wy - targets_temp[camerai]->group.wy)/(targets_temp[camerai]->minDot.wx - targets_temp[camerai]->group.wx));
								targets_temp[camerai]->angle = 180 + targets_temp[camerai]->angle*180.0/3.14159;
							}
							else												//quadrant 2
							{
								targets_temp[camerai]->angle = atan((targets_temp[camerai]->minDot.wy - targets_temp[camerai]->group.wy)/(targets_temp[camerai]->minDot.wx - targets_temp[camerai]->group.wx));
								targets_temp[camerai]->angle = 180 + targets_temp[camerai]->angle*180.0/3.14159;
							}
						}
						else
						{
							if(targets_temp[camerai]->minDot.wy < targets_temp[camerai]->group.wy)				//quadrant 4
							{
								targets_temp[camerai]->angle = atan((targets_temp[camerai]->minDot.wy - targets_temp[camerai]->group.wy)/(targets_temp[camerai]->minDot.wx - targets_temp[camerai]->group.wx));
								targets_temp[camerai]->angle = targets_temp[camerai]->angle*180.0/3.14159;
							}
							else												//quandrant 1
							{
								targets_temp[camerai]->angle = atan((targets_temp[camerai]->minDot.wy - targets_temp[camerai]->group.wy)/(targets_temp[camerai]->minDot.wx - targets_temp[camerai]->group.wx));
								targets_temp[camerai]->angle = targets_temp[camerai]->angle*180.0/3.14159;
							}
						}
						/*cout<<"("<<targets_temp[camerai]->minDot.wx<<","<<targets_temp[camerai]->minDot.wy<<") ";
						cout<<"("<<targets_temp[camerai]->group.wx<<","<<targets_temp[camerai]->group.wy<<") ";
						cout<<"offset: "<<targets_temp[camerai]->angleoffset;
						cout<<"angle: "<<targets_temp[camerai]->angle<<endl;
						*/

						targets_temp[camerai]->angle -= targets_temp[camerai]->angleoffset;
						while(targets_temp[camerai]->angle < 0)
						{
							targets_temp[camerai]->angle += 360;
						}
						targets_temp[camerai]->angle = (int) (targets_temp[camerai]->angle) % 360;
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}
					
					//where file writing was			

					//plot centers of mass and angle
					
					/*targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						cvLine(world_image, cvPoint((int) targets_temp[camerai]->minDot.ix, (int) targets_temp[camerai]->minDot.iy), cvPoint((int) targets_temp[camerai]->group.ix, (int) targets_temp[camerai]->group.iy), cvScalar(0,255,0), 2);
						cvCircle(world_image, cvPoint((int) targets_temp[camerai]->group.ix, (int) targets_temp[camerai]->group.iy), 3, cvScalar(0, 0, 255));
						cvRectangle(world_image, cvPoint((int) targets_temp[camerai]->x1, (int) targets_temp[camerai]->y1), cvPoint((int) targets_temp[camerai]->x2, (int) targets_temp[camerai]->y2), cvScalar(0, 255, 0));
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}

					//plot text
					CvFont font;
					double hScale=0.5;
					double vScale=0.5;
					int    lineWidth=1;
					cvInitFont(&font,CV_FONT_HERSHEY_SIMPLEX|CV_FONT_ITALIC, hScale,vScale,0,lineWidth);
					
					targets_temp[camerai] = first_target;
					while(targets_temp[camerai])
					{
						string output_string;
						output_string = "(";
						ostringstream oss;
						oss.str("");
						oss<<targets_temp[camerai]->ID;
						string s;
						s = oss.str();
						output_string.append(s);
						output_string.append(",");
						oss.str("");
						oss<<targets_temp[camerai]->group.wx;
						s = oss.str();
						output_string.append(s.substr(0,4));
						output_string.append(",");
						oss.str("");
						oss<<targets_temp[camerai]->group.wy;
						s = oss.str();
						output_string.append(s.substr(0,4));
						output_string.append(",");
						oss.str("");
						oss<<targets_temp[camerai]->angle;
						s = oss.str();
						output_string.append(s.substr(0,3));
						output_string.append(")");
						cvPutText (world_image,output_string.c_str(),cvPoint((int) targets_temp[camerai]->group.ix + 5,(int) targets_temp[camerai]->group.iy - 15), &font, CV_RGB(255,0,0));
						targets_temp[camerai] = targets_temp[camerai]->nxt;
					}
					
					*/

					targets_temp[camerai] = first_target;

					//output resultant.
					
					//cvShowImage( "TrackSys2", world_image);

					// video stream writing turned off for increased real time speed
					//cvWriteFrame(writer1, world_image);
					//cvWriteFrame(writer2, image);
					

				}
			
			}
		}

		// ALL 4 cameras are processed
		// Time to output resultant captured data

		//Calculate FPS
		// (note this is a poor calculation, should use a moving window average
		numFrames++;
		long newtime = clock();
		fps = (1.0 * numFrames) / ((1.0*newtime - startTime)/1000) ;
		
		// total count of patterns from 4 cameras
		int t_count = 0;
		
		//******************************************************//
		//********PHASE 3c: BLIND SPOT REJECTION****************//
		//******************************************************//	
		// used to reject multiple patterns located or partials.

		// process each cameras linkedlist of targets
		for (int camerai = 0; camerai<4; camerai++)
		{
			//initialize head of LL
			target* temp = targets_temp[camerai];
			
			// process entire LL for this head
			while(temp)
			{
				if(temp->ID != -1)
				{
					// Check every other target from all other cameras
					for (int camerai2 = 0; camerai2<4; camerai2++)
					{
						target* temp2 = targets_temp[camerai2];
				
						while(temp2)	// if more targets in linked list
						{
							if(temp2->ID != -1)		// if the ID is a legitmate target (! -1)
							{
								if ((temp2 != temp) || (camerai != camerai2))	//ensure we're not comparing target to itself
								{
									// Calculate distance between patterns
									double temp_dist = sqrt(pow(temp->group.wx - temp2->group.wx,2) + pow( (temp->group.wy  - temp2->group.wy),2));

									//same pattern recognized
									// both cameras saw full images of the targets and their IDs match
									if (temp->ID == temp2->ID)
									{
										//average parameters (x and y coordintes)
										temp->group.wx = (temp->group.wx + temp2->group.wx) / 2;
										temp->group.wy = (temp->group.wy + temp2->group.wy) / 2;
										//reset temp2
										temp2->ID = -1;
										
									}
									else if (temp_dist < max_dot_distance)
									{
										// different pattern, but within range, so accept pattern with more dots visible
										if (temp->numDots > temp2->numDots)
										{
											temp2->ID = -1;
										}
										else
										{
											temp->ID = -1;
										}
									}
								}
								
							}
							temp2 = temp2->nxt;
						}
						
					}


					//update total count
					t_count++;				

					//output to console
					//cout<<"Camera "<<cameraID[camerai]<<" Object "<<t_count<<": ("<<temp->ID<<","<<temp->group.wx<<","<<temp->group.wy<<","<<temp->angle<<")"<<endl;
					//cout<<"Camera "<<cameraID[camerai]<<" Object "<<t_count<<": ("<<temp->ID<<","<<temp->group.wx<<","<<temp->group.wy<<","<<temp->angle<<")"<<endl;
					fprintf(stdout,"%8d  %12.3f  %12.3f  %12.3f\n",temp->ID, temp->group.wx, temp->group.wy, temp->angle);

					//output to filestream
					fOut<<newtime<<","<<temp->ID<<","<<temp->group.wx<<","<<temp->group.wy<<","<<temp->angle<<endl;

					//output to serial port
					ostringstream oss;
					string serialData;
					oss.str("");
					oss<<temp->ID<<","<<temp->group.wx<<","<<temp->group.wy<<","<<temp->angle<<endl;
					serialData = oss.str();
					serialout.Write(serialData.c_str());
				}
				
				temp = temp->nxt;
			}
			
		}

		//Update FPS and target data for GUI
		message1 = "Real-Time Localization Active. Press Esc to Quit";
		ostringstream oss;
		oss.str("");
		oss<<t_count<<" targets at "<<fps<<" fps.";
		message2 = oss.str();
		
		//delete linked list of targets
		for(int camerai = 0; camerai<4; camerai++)
		{
			target* temp1 = targets_temp[camerai];
			target* temp_old;		
			while(temp1)
			{
				temp_old = temp1;
				temp1 = temp1->nxt;
				delete temp_old;
			}
		}

		//CHECK KEYSTROKES
		if(process_keys()==1)
		{
			break;
		}

	}

	// Get here when user presses escape, otherwise this loops back up to main program loop


	//END OF PROGRAM, FREE UP RESOURCES

	//REALEASE CAMERAS
	for(int camerai = 0; camerai<4; camerai++)
	{
		VI->stopDevice(cameraID[camerai]);
	}

	//RELEASE FILESTREAM 
	fOut.close();
	
	//RELEASE IMAGES
	for(int camerai = 0; camerai<4; camerai++)
	{
		cvReleaseImage(&greyImage[camerai]);
		cvReleaseImage(&thresholdedImageColor[camerai]);
		cvReleaseImage(&thresholdedImage[camerai]);
	}
	cvReleaseImage(&image);
	cvReleaseImage(&grey);
	cvReleaseImage(&thresholded_image);
	cvReleaseMemStorage( &storage);
    cvDestroyWindow(mainWindow.c_str());
	cvDestroyWindow(vidWindow.c_str());
	delete(VI);
	exit(333);

    return 0;
}

#ifdef _EiC
main(1,"lkdemo.c");
#endif